var zodToJsonSchema = require('zod-to-json-schema');

class OpenAPISpecification {
  static isValidOpenApiVersion(version = '') {
    const match = /(\d+)\.(\d+).(\d+)/.exec(version);

    if (match) {
      const major = Number(match[1]);

      if (major >= 3) {
        return true;
      }
    }

    return false;
  }

  assertNonNull(value, name) {
    if (!value) {
      throw new Error(`${name} is not set!`);
    }
  }

  constructor({
    openapi = '3.0.0',
    info = {
      title: '',
      version: ''
    },
    components,
    paths = {},
    security = [],
    tags = [],
    servers = [],
    externalDocs
  } = {
    openapi: '3.0.0',
    info: {
      title: '',
      version: ''
    },
    paths: {}
  }) {
    this.spec = void 0;
    this.spec = {
      openapi,
      info,
      components: {
        schemas: {},
        responses: {},
        parameters: {},
        examples: {},
        requestBodies: {},
        headers: {},
        securitySchemes: {},
        links: {},
        callbacks: {},
        ...components
      },
      paths,
      security,
      tags,
      servers,
      externalDocs
    };
    this.setOpenApiVersion(openapi);
  }

  get openapi() {
    return this.spec.openapi;
  }

  setOpenApiVersion(version) {
    if (!OpenAPISpecification.isValidOpenApiVersion(version)) {
      throw new Error(`Invalid OpenAPI version: ${version}, expected format: 3.y.z`);
    }

    this.spec.openapi = version;
    return this;
  }

  get info() {
    return this.spec.info;
  }

  setInfo(info) {
    this.spec.info = info;
    return this;
  }

  get components() {
    return this.spec.components;
  }

  setComponent(componentType, name, component) {
    this.spec.components[componentType] = { ...this.spec.components[componentType],
      [name]: component
    };
    return this;
  }

  get paths() {
    return this.spec.paths;
  }

  getPathItem(path) {
    var _this$paths$path;

    return (_this$paths$path = this.paths[path]) != null ? _this$paths$path : null;
  }

  setPathItem(path, pathItem) {
    this.spec.paths[path] = pathItem;
    return this;
  }

  addPathItemParameter(path, parameter) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);

    if (!pathItem.parameters) {
      pathItem.parameters = [];
    }

    pathItem.parameters.push(parameter);
    this.spec.paths[path] = pathItem;
    return this;
  }

  addPathItemServer(path, server) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);

    if (!pathItem.servers) {
      pathItem.servers = [];
    }

    pathItem.servers.push(server);
    this.spec.paths[path] = pathItem;
    return this;
  }

  getPathItemOperation(path, method) {
    var _this$paths$path$meth, _this$paths$path2;

    return (_this$paths$path$meth = (_this$paths$path2 = this.paths[path]) == null ? void 0 : _this$paths$path2[method]) != null ? _this$paths$path$meth : null;
  }

  setPathItemOperation(path, method, operation) {
    const pathItem = this.getPathItem(path);
    this.assertNonNull(pathItem, `paths[${path}] pathItem`);
    this.spec.paths[path] = { ...pathItem,
      [method]: operation
    };
    return this;
  }

  setPathItemOperationRequestBody(path, method, requestBody) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.requestBody = requestBody;
    return this;
  }

  getPathItemOperationResponse(path, method, httpStatusCode) {
    var _this$paths$path$meth2, _this$paths$path3, _this$paths$path3$met, _this$paths$path3$met2;

    return (_this$paths$path$meth2 = (_this$paths$path3 = this.paths[path]) == null ? void 0 : (_this$paths$path3$met = _this$paths$path3[method]) == null ? void 0 : (_this$paths$path3$met2 = _this$paths$path3$met.responses) == null ? void 0 : _this$paths$path3$met2[httpStatusCode]) != null ? _this$paths$path$meth2 : null;
  }

  setPathItemOperationResponse(path, method, httpStatusCode, response) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.responses[httpStatusCode] = response;
    return this.setPathItemOperation(path, method, operation);
  }

  addPathItemOperationTag(path, method, tag) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.tags = operation.tags ? operation.tags.concat(tag) : [tag];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationParameter(path, method, parameter) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.parameters = operation.parameters ? operation.parameters.concat(parameter) : [parameter];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationSecurityRequirement(path, method, securityRequirement) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.security = operation.security ? operation.security.concat(securityRequirement) : [securityRequirement];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  addPathItemOperationServer(path, method, server) {
    const operation = this.getPathItemOperation(path, method);
    this.assertNonNull(operation, `paths[${path}][${method}] operation`);
    operation.servers = operation.servers ? operation.servers.concat(server) : [server];
    this.setPathItemOperation(path, method, operation);
    return this;
  }

  get security() {
    return this.spec.security;
  }

  addSecurity(security) {
    this.spec.security.push(security);
    return this;
  }

  get tags() {
    return this.spec.tags;
  }

  addTag(tag) {
    this.spec.tags.push(tag);
    return this;
  }

  get servers() {
    return this.spec.servers;
  }

  addServer(server) {
    this.spec.servers.push(server);
    return this;
  }

  get externalDocs() {
    return this.spec.externalDocs;
  }

  setExternalDocs(externalDocs) {
    this.spec.externalDocs = externalDocs;
    return this;
  }

  toJSON() {
    return this.spec;
  }

  toString(replacer, space) {
    return JSON.stringify(this.spec, replacer, space);
  }

}

const regexPattern = {
  param: /\(\?:\(\[\^\\\/]\+\?\)\)/,
  path: /^\^\\\/(?:(:?[\w\\.-]*(?:\\\/:?[\w\\.-]*)*)|(\(\?:\(\[\^\\\/]\+\?\)\)))\\/,
  rootPath: /^\^\\\/\?\(\?=\\\/\|\$\)$/
};

const getPathString = (pathRegex, keys) => {
  var _regexStringExecArray;

  const pathRegexpSource = pathRegex.source;
  let regexString = pathRegexpSource;
  let regexStringExecArray = regexPattern.path.exec(regexString);
  let paramIndex = 0;

  while (regexPattern.param.test(regexString)) {
    const param = `:${keys[paramIndex].name}`;
    regexString = regexString.replace(regexPattern.param, param);
    paramIndex++;
  }

  if (regexString !== pathRegexpSource) {
    regexStringExecArray = regexPattern.path.exec(regexString);
  }

  const pathString = (_regexStringExecArray = regexStringExecArray) == null ? void 0 : _regexStringExecArray[1].replace(/\\\//g, '/');
  return pathString != null ? pathString : null;
};

const processRoute = (spec, route, basePath = '', plugins) => {
  const routePaths = Array.isArray(route.path) ? route.path : [route.path];
  routePaths.forEach(routePath => {
    const path = `${basePath ? `${basePath}${routePath === '/' ? '' : routePath}` : routePath}`.replace(/:([a-zA-Z0-9_]+)/g, '{$1}');
    const methods = Object.keys(route.methods).filter(method => method !== '_all');

    if (!spec.paths[path]) {
      spec.setPathItem(path, {});
    }

    for (const method of methods) {
      spec.setPathItemOperation(path, method, {
        responses: {
          default: {
            description: ''
          }
        }
      });

      for (const plugin of plugins) {
        const stashedValue = plugin.stash.find(route, method);

        if (stashedValue) {
          plugin.processRoute(plugin.specification, stashedValue, {
            path,
            method
          });
        }
      }
    }
  });
};

const processExpressRouters = (spec, app, basePath = '', plugins) => {
  var _app$stack, _app$_router;

  const stack = (_app$stack = app.stack) != null ? _app$stack : (_app$_router = app._router) == null ? void 0 : _app$_router.stack;

  if (!stack) {
    spec.setPathItem(basePath, {});
    return;
  }

  for (const layer of stack) {
    // terminal route
    if (layer.route) {
      processRoute(spec, layer.route, basePath, plugins);
      continue;
    }

    if (layer.name !== "router"
    /* StackName.ROUTER */
    && layer.name !== "bound dispatch"
    /* StackName.BOUND_DISPATCH */
    && layer.name !== "mounted_app"
    /* StackName.MOUNTED_APP */
    ) {
      continue;
    }

    if (regexPattern.path.test(layer.regexp.source)) {
      const parsedPath = getPathString(layer.regexp, layer.keys);

      if (parsedPath) {
        processExpressRouters(spec, layer.handle, basePath + '/' + parsedPath, plugins);
      }

      continue;
    }

    if (!layer.path && layer.regexp && !regexPattern.rootPath.test(layer.regexp.source)) {
      const regexPath = ' RegExp(' + layer.regexp + ') ';
      processExpressRouters(spec, layer.handle, basePath + '/' + regexPath, plugins);
      continue;
    }

    processExpressRouters(spec, layer.handle, basePath, plugins);
  }
};

class Stash {
  constructor(stashKey) {
    this.stashKey = void 0;

    this.find = (route, method) => {
      for (const layer of route.stack) {
        if (layer.method === method) {
          const descriptor = Object.getOwnPropertyDescriptor(layer.handle, this.stashKey);

          if (descriptor) {
            return descriptor.value;
          }
        }
      }

      return null;
    };

    this.store = (handler, value) => {
      Object.defineProperty(handler, this.stashKey, {
        value,
        enumerable: true
      });
    };

    this.stashKey = stashKey;
  }

}

class RequestValidationError extends Error {
  constructor(validationError, segment) {
    super(validationError.message);
    this.segment = void 0;
    this.validationError = void 0;
    this.segment = segment;
    this.validationError = validationError;
  }

}
class ResponseValidationError extends Error {
  constructor(validationError, segment) {
    super(validationError.message);
    this.segment = void 0;
    this.validationError = void 0;
    this.segment = segment;
    this.validationError = validationError;
  }

}

async function validateRequest(req, spec, segmentOrder) {
  await segmentOrder.reduce((promise, segment) => {
    return promise.then(async () => {
      const schema = spec.get(segment);

      if (!schema) {
        return null;
      }

      const result = await schema.safeParseAsync(req[segment]);

      if (!result.success) {
        throw new RequestValidationError(result.error, segment);
      }

      req[segment] = result.data;
      return null;
    });
  }, Promise.resolve(null));
}
function setupResponseValidation(res, spec, segmentOrder, next) {
  const originalJson = res.json;

  res.json = function validateAndSendJsonResponse(...args) {
    res.json = originalJson;
    const value = {
      body: args[0],
      headers: res.getHeaders()
    };
    const schemaBySegment = spec.has(String(res.statusCode)) ? spec.get(String(res.statusCode)) : spec.get('default');

    if (!schemaBySegment) {
      next(new Error(`Validation Schema not found for Response(${res.statusCode})`));
      return res;
    }

    for (const segment of segmentOrder) {
      const schema = schemaBySegment[segment];

      if (!schema) {
        continue;
      }

      const result = schema.safeParse(value[segment]);

      if (!result.success) {
        next(new ResponseValidationError(result.error, segment));
        return res;
      }

      if (segment === 'body') {
        value[segment] = result.data;
      }
    }

    return res.json(value.body);
  };
}

const parameterLocationBySegment = {
  cookies: 'cookie',
  signedCookies: 'cookie',
  headers: 'header',
  params: 'path',
  query: 'query'
};

const routeProcessor = (specification, specInfo, {
  path,
  method
}) => {
  for (const [segment, schema] of specInfo.req.entries()) {
    if (!schema) {
      continue;
    }

    if (specInfo.operationId) {
      specification.setPathItemOperation(path, method, { ...specification.getPathItemOperation(path, method),
        operationId: specInfo.operationId
      });
    }

    if (segment === 'body') {
      const result = zodToJsonSchema.zodToJsonSchema(schema);
      const requestBody = {
        content: {
          'application/json': {
            // @ts-expect-error ...
            schema: result
          }
        }
      };
      specification.setPathItemOperationRequestBody(path, method, requestBody);
      continue;
    }

    const location = parameterLocationBySegment[segment];
    const result = zodToJsonSchema.zodToJsonSchema(schema);

    if ('properties' in result) {
      const {
        properties,
        required = []
      } = result;

      if (properties) {
        for (const [name, schema] of Object.entries(properties)) {
          const parameterObject = {
            name,
            in: location,
            // @ts-expect-error ...
            schema: schema,
            required: required.includes(name)
          };
          specification.addPathItemOperationParameter(path, method, parameterObject);
        }
      }
    }
  }

  for (const [key, schemaBySegment] of specInfo.res.entries()) {
    if (!schemaBySegment) {
      continue;
    }

    const httpStatusCode = key;

    if (schemaBySegment.body) {
      const result = zodToJsonSchema.zodToJsonSchema(schemaBySegment.body);
      const responseObject = { ...specification.getPathItemOperationResponse(path, method, httpStatusCode),
        description: '',
        content: {
          'application/json': {
            // @ts-expect-error ...
            schema: result
          }
        }
      };
      specification.setPathItemOperationResponse(path, method, httpStatusCode, responseObject);
    }

    if (schemaBySegment.headers) {
      const result = zodToJsonSchema.zodToJsonSchema(schemaBySegment.headers);

      if ('properties' in result) {
        const {
          properties,
          required = []
        } = result;
        const headersObject = {};

        if (properties) {
          for (const [name, schema] of Object.entries(properties)) {
            const headerObject = {
              // @ts-expect-error ...
              schema: schema,
              required: required.includes(name)
            };
            headersObject[name] = headerObject;
          }
        }

        const responseObject = { ...specification.getPathItemOperationResponse(path, method, httpStatusCode),
          description: '',
          headers: headersObject
        };
        specification.setPathItemOperationResponse(path, method, httpStatusCode, responseObject);
      }
    }
  }
};

const create = ({
  req: reqConfig = {},
  res: resConfig = {}
} = {}) => {
  const {
    segmentOrder: reqSegmentOrder = ['headers', 'params', 'query', 'cookies', 'signedCookies', 'body'],
    skipValidation: skipRequestValidation = false
  } = reqConfig;
  const {
    segmentOrder: resSegmentOrder = ['body', 'headers'],
    skipValidation: skipResponseValidation = false
  } = resConfig;
  const specificationPlugin = {
    name: 'zod-openapi-spec',
    getMiddleware: (internals, validationSchema) => {
      const specInfo = {
        operationId: validationSchema.operationId,
        req: new Map(),
        res: new Map()
      };

      for (const segment of reqSegmentOrder) {
        var _validationSchema$req;

        const schema = (_validationSchema$req = validationSchema.req) == null ? void 0 : _validationSchema$req[segment];

        if (schema) {
          specInfo.req.set(segment, schema);
        }
      }

      for (const [code, schemaBySegment] of Object.entries((_validationSchema$res = validationSchema.res) != null ? _validationSchema$res : {})) {
        var _validationSchema$res;

        specInfo.res.set(code, Object.entries(schemaBySegment).reduce((bySegment, [segment, schema]) => {
          if (schema) {
            bySegment[segment] = schema;
          }

          return bySegment;
        }, {}));
      }

      const validationMiddleware = async (req, res, next) => {
        if (!skipResponseValidation && validationSchema.res) {
          setupResponseValidation(res, specInfo.res, resSegmentOrder, next);
        }

        if (!skipRequestValidation && validationSchema.req) {
          try {
            await validateRequest(req, specInfo.req, reqSegmentOrder);
          } catch (err) {
            next(err);
            return;
          }
        }

        next();
      };

      internals.stash.store(validationMiddleware, specInfo);
      return validationMiddleware;
    },
    processRoute: routeProcessor
  };
  return specificationPlugin;
};
const SpecificationPlugin = {
  create,
  routeProcessor
};

class ExpressOpenAPI {
  constructor({
    specification = new OpenAPISpecification(),
    useStringStashKey = false
  } = {}) {
    this.plugins = [];
    this.useStringStashKey = void 0;
    this.specification = void 0;

    this.registerPlugin = plugin => {
      const stashSymbol = Symbol(plugin.name);
      const stashKey = this.useStringStashKey ? stashSymbol.toString() : stashSymbol;
      const internals = {
        specification: this.specification,
        stash: new Stash(stashKey)
      };

      const getMiddleware = (...params) => {
        return plugin.getMiddleware(internals, ...params);
      };

      const registeredPlugin = {
        name: plugin.name,
        specification: internals.specification,
        stash: internals.stash,
        processRoute: plugin.processRoute
      };
      this.plugins.push(registeredPlugin);
      return getMiddleware;
    };

    this.populateSpecification = (app, basePath) => {
      processExpressRouters(this.specification, app, basePath, this.plugins);
      return this.specification;
    };

    this.specification = specification;
    this.useStringStashKey = useStringStashKey;
  }

}

exports.ExpressOpenAPI = ExpressOpenAPI;
exports.OpenAPISpecification = OpenAPISpecification;
exports.RequestValidationError = RequestValidationError;
exports.ResponseValidationError = ResponseValidationError;
exports.SpecificationPlugin = SpecificationPlugin;
exports.setupResponseValidation = setupResponseValidation;
exports.validateRequest = validateRequest;
//# sourceMappingURL=index.js.map
